package com.sun.org.apache.xerces.internal.impl.xs.identity;

import com.sun.org.apache.xerces.internal.xs.XSComplexTypeDefinition;
import com.sun.org.apache.xerces.internal.xs.XSTypeDefinition;
import com.sun.org.apache.xerces.internal.impl.xs.util.ShortListImpl;
import com.sun.org.apache.xerces.internal.xs.ShortList;
import com.sun.org.apache.xerces.internal.util.XMLChar;
import com.sun.org.apache.xerces.internal.impl.xpath.XPathException;
import com.sun.org.apache.xerces.internal.xni.NamespaceContext;
import com.sun.org.apache.xerces.internal.util.SymbolTable;
import com.sun.org.apache.xerces.internal.impl.xpath.XPath;

public class Field
{
    protected final Field.XPath fXPath;
    protected final IdentityConstraint fIdentityConstraint;
    
    public Field(Field.XPath xpath, IdentityConstraint identityConstraint) {
        this.fXPath = xpath;
        this.fIdentityConstraint = identityConstraint;
    }
    
    public com.sun.org.apache.xerces.internal.impl.xpath.XPath getXPath() {
        return this.fXPath;
    }
    
    public IdentityConstraint getIdentityConstraint() {
        return this.fIdentityConstraint;
    }
    
    public XPathMatcher createMatcher(ValueStore store) {
        return new Field.Matcher(this.fXPath, store);
    }
    
    @Override
    public String toString() {
        return this.fXPath.toString();
    }
    
    public static class XPath extends com.sun.org.apache.xerces.internal.impl.xpath.XPath
    {
        public XPath(String xpath, SymbolTable symbolTable, NamespaceContext context) throws XPathException {
            super(fixupXPath(xpath), symbolTable, context);
            for (int i = 0; i < this.fLocationPaths.length; ++i) {
                int j = 0;
                while (j < this.fLocationPaths[i].steps.length) {
                    XPath.Axis axis = this.fLocationPaths[i].steps[j].axis;
                    if (axis.type == 2 && j < this.fLocationPaths[i].steps.length - 1) {
                        throw new XPathException("c-fields-xpaths");
                    }
                    ++j;
                }
            }
        }
        
        private static String fixupXPath(String xpath) {
            int end = xpath.length();
            int offset = 0;
            boolean whitespace = true;
            while (offset < end) {
                char c = xpath.charAt(offset);
                if (whitespace) {
                    if (!XMLChar.isSpace((int)c)) {
                        if (c == '.' || c == '/') {
                            whitespace = false;
                        }
                        else if (c != '|') {
                            return fixupXPath2(xpath, offset, end);
                        }
                    }
                }
                else if (c == '|') {
                    whitespace = true;
                }
                ++offset;
            }
            return xpath;
        }
        
        private static String fixupXPath2(String xpath, int offset, int end) {
            StringBuffer buffer = new StringBuffer(end + 2);
            for (int i = 0; i < offset; ++i) {
                buffer.append(xpath.charAt(i));
            }
            buffer.append("./");
            boolean whitespace = false;
            while (offset < end) {
                char c = xpath.charAt(offset);
                if (whitespace) {
                    if (!XMLChar.isSpace((int)c)) {
                        if (c == '.' || c == '/') {
                            whitespace = false;
                        }
                        else if (c != '|') {
                            buffer.append("./");
                            whitespace = false;
                        }
                    }
                }
                else if (c == '|') {
                    whitespace = true;
                }
                buffer.append(c);
                ++offset;
            }
            return buffer.toString();
        }
    }
    
    protected class Matcher extends XPathMatcher
    {
        protected final ValueStore fStore;
        protected boolean fMayMatch;
        
        public Matcher(Field.XPath xpath, ValueStore store) {
            super(xpath);
            this.fMayMatch = true;
            this.fStore = store;
        }
        
        @Override
        protected void matched(Object actualValue, short valueType, ShortList itemValueType, boolean isNil) {
            super.matched(actualValue, valueType, itemValueType, isNil);
            if (isNil && Field.this.fIdentityConstraint.getCategory() == 1) {
                String code = "KeyMatchesNillable";
                this.fStore.reportError(code, new Object[] { Field.this.fIdentityConstraint.getElementName(), Field.this.fIdentityConstraint.getIdentityConstraintName() });
            }
            this.fStore.addValue(Field.this, this.fMayMatch, actualValue, this.convertToPrimitiveKind(valueType), this.convertToPrimitiveKind(itemValueType));
            this.fMayMatch = false;
        }
        
        private short convertToPrimitiveKind(short valueType) {
            if (valueType <= 20) {
                return valueType;
            }
            if (valueType <= 29) {
                return 2;
            }
            if (valueType <= 42) {
                return 4;
            }
            return valueType;
        }
        
        private ShortList convertToPrimitiveKind(ShortList itemValueType) {
            if (itemValueType != null) {
                int length = itemValueType.getLength();
                int i = 0;
                while (i < length) {
                    short type = itemValueType.item(i);
                    if (type != this.convertToPrimitiveKind(type)) {
                        break;
                    }
                    ++i;
                }
                if (i != length) {
                    short[] arr = new short[length];
                    for (int j = 0; j < i; ++j) {
                        arr[j] = itemValueType.item(j);
                    }
                    while (i < length) {
                        arr[i] = this.convertToPrimitiveKind(itemValueType.item(i));
                        ++i;
                    }
                    return new ShortListImpl(arr, arr.length);
                }
            }
            return itemValueType;
        }
        
        @Override
        protected void handleContent(XSTypeDefinition type, boolean nillable, Object actualValue, short valueType, ShortList itemValueType) {
            if (type == null || (type.getTypeCategory() == 15 && ((XSComplexTypeDefinition)type).getContentType() != 1)) {
                this.fStore.reportError("cvc-id.3", new Object[] { Field.this.fIdentityConstraint.getName(), Field.this.fIdentityConstraint.getElementName() });
            }
            this.matched(this.fMatchedString = actualValue, valueType, itemValueType, nillable);
        }
    }
}
